#include <stdio.h>
#include <errno.h>
#include <ctype.h>

#include "jistbl.h"

#define BUFLEN 256

void
ucs4_to_utf8(int d)
{
    if (d == 9825) {
	printf("&#9825;");
	return;
    }
    if (d <= 0x7f) {
	putchar(d);
    } else if (d <= 0x7ff) {
	putchar(((d >> 6) & 0x1f) | 0xc0);
	putchar((d & 0x3f) | 0x80);
    }
    putchar(((d >> 12) & 0x0f) | 0xe0);
    putchar((d >> 6) & 0x3f | 0x80);
    putchar((d & 0x3f) | 0x80);
}

void
convert(FILE* fp)
{
    int c, d;
    int idx;
    int hex;
    char buf[BUFLEN];
    
    while (!feof(fp)) {
	if ((c = fgetc(fp)) == EOF)
	    break;
	if (c != '&') {
	    putchar(c);
	    continue;
	}
	
	if ((c = fgetc(fp)) == EOF) {
	    putchar('&');
	    break;
	}
	if (c != '#') {
	    printf("&%c", c);
	    continue;
	}
	idx = 0;
	d = 0;
	hex = 0;
	while (!feof(fp) && idx < BUFLEN - 1) {
	    if ((c = fgetc(fp)) == EOF)
		break;
	    buf[idx++] = c;
	    if (isdigit(c)) {
		d = d * 10 + (c - '0');
	    } else {
		break;
	    }
	}
	if (!feof(fp) && c == ';') {
	    if (ucs_to_jis[d]) {
		ucs4_to_utf8(d);
	    } else {
		printf("&#%u;", d);
	    }
	} else {
	    buf[idx] = '\0';
	    printf("&#%s;", buf);
	}
    }
}

void
usage()
{
    fprintf(stderr, "usage: charunref file : read from file\n       charunref -    : read from stdin\n");
    exit (1);
}

int
main(int argc, char* argv[])
{
    if (argc != 2 || !strcmp(argv[1], "-h")) {
	usage();
    } else {
	FILE* fp = strcmp(argv[1], "-") ? fopen(argv[1], "r") : stdin;
	if (!fp) {
	    fprintf(stderr, "%s: Error: %s\n", argv[1], strerror(errno));
	    exit(1);
	}
	convert(fp);
	if (strcmp(argv[1], "-")) fclose(fp);
    }
    return 0;
}

/* end of charunref.c */
